"""Provides visualization utilities."""

import logging
import pathlib
from typing import Optional

import matplotlib as mpl
import matplotlib.pyplot as plt


log = logging.getLogger(__name__)


def _save_and_close(
    fig: mpl.figure.Figure, path: Optional[pathlib.Path], close_figure: bool
) -> None:
    """Saves a figure at a given location if path is provided and optionally closes it.

    Args:
        fig (matplotlib.figure.Figure): figure to save
        path (Optional[pathlib.Path]): path where figure should be saved, or None to not
            save it
        close_figure (bool): whether to close figure after saving
    """
    if path is not None:
        path.parent.mkdir(parents=True, exist_ok=True)
        log.info(f"saving figure as {path}")
        fig.savefig(path)
    if close_figure:
        plt.close(fig)


def _log_figure_path(path: Optional[pathlib.Path]) -> Optional[pathlib.Path]:
    """Adds a suffix to a figure path to indicate the use of a logarithmic axis.

    If the path is None (since figure should not be saved), it will stay None.

    Args:
        path (Optional[pathlib.Path]): original path to figure, or None if figure should
            not be saved

    Returns:
        Optional[pathlib.Path]: new path to figure including _log suffix, or None if
            original path is None
    """
    if path is not None:
        return path.with_name(path.stem + "_log" + path.suffix)
    return None



"""High-level entry point for visualizing fit models and inference results."""

import glob
import logging
import pathlib
from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import pyhf

from cabinetry import configuration
from cabinetry import fit
from cabinetry import histo
from cabinetry import model_utils
from cabinetry.templates import builder
from cabinetry.visualize import plot_model
from cabinetry.visualize import plot_result


log = logging.getLogger(__name__)


def _figure_name(region_name: str, label: str) -> str:
    """Constructs a file name for a figure.

    Args:
        region_name (str): name of the region shown in the figure
        label (str): additional label, e.g. from model prediction

    Returns:
        str: name of the file the figure should be saved to
    """
    if label in ["pre-fit", "post-fit"]:
        label = label.replace("-", "")  # replace pre-fit by prefit, post-fit by postfit
    figure_name = f"{region_name.replace(' ', '-')}_{label}.pdf"
    return figure_name


def _total_yield_uncertainty(stdev_list: List[np.ndarray]) -> np.ndarray:
    """Calculates the absolute statistical uncertainty of a stack of MC.

    Args:
        stdev_list (List[np.ndarray]): list of absolute stat. uncertainty per sample

    Returns:
        np.ndarray: absolute stat. uncertainty of stack of samples
    """
    tot_unc = np.sqrt(np.sum(np.power(np.asarray(stdev_list), 2), axis=0))
    return tot_unc


def data_mc_from_histograms(
    config: Dict[str, Any],
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    log_scale: Optional[bool] = None,
    log_scale_x: bool = False,
    channels: Optional[Union[str, List[str]]] = None,
    colors: Optional[Dict[str, str]] = None,
    close_figure: bool = False,
    save_figure: bool = True,
) -> List[Dict[str, Any]]:
    """Draws pre-fit data/MC histograms, using histograms created by cabinetry.

    The uncertainty band drawn includes only statistical uncertainties.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        log_scale (Optional[bool], optional): whether to use logarithmic vertical axis,
            defaults to None (automatically determine whether to use linear/log scale)
        log_scale_x (bool, optional): whether to use logarithmic horizontal axis,
            defaults to False
        channels (Optional[Union[str, List[str]]], optional): name of channel to show,
            or list of names to include, defaults to None (uses all channels)
        colors (Optional[Dict[str, str]], optional): map of sample names and colors to
            use in plot, defaults to None (uses default colors)
        close_figure (bool, optional): whether to close each figure, defaults to False
            (enable when producing many figures to avoid memory issues, prevents
            automatic rendering in notebooks)
        save_figure (bool, optional): whether to save figures, defaults to True

    Raises:
        ValueError: if color specification is incomplete

    Returns:
        List[Dict[str, Any]]: list of dictionaries, where each dictionary contains a
            figure and the associated region name
    """
    log.info("visualizing histogram")
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    figure_dict_list = []

    # if custom colors are provided, ensure that they cover all samples
    if colors is not None:
        c_missing = {
            sample["Name"] for sample in config["Samples"] if not sample.get("Data")
        }.difference(colors.keys())
        if c_missing:
            raise ValueError(
                f"colors need to be provided for all samples, missing for {c_missing}"
            )

    # create a list of channels to process
    if channels is not None and isinstance(channels, str):
        channels = [channels]

    for region in config["Regions"]:
        if channels is not None and region["Name"] not in channels:
            continue  # skip region
        histogram_dict_list = []
        model_stdevs = []
        # loop over samples in reverse order, such that samples that appear first in the
        # config will appear at the top of the stack in the plot (they are plotted last)
        for sample in config["Samples"][::-1]:
            is_data = sample.get("Data", False)
            histogram = histo.Histogram.from_config(
                histogram_folder, region, sample, {}, modified=True
            )
            histogram_dict_list.append(
                {
                    "label": sample["Name"],
                    "isData": is_data,
                    "yields": histogram.yields,
                    "variable": region["Variable"],
                }
            )
            if not is_data:
                model_stdevs.append(histogram.stdev)

        figure_name = _figure_name(region["Name"], "pre-fit")
        total_model_unc = _total_yield_uncertainty(model_stdevs)
        bin_edges = histogram.bins
        label = f"{region['Name']}\npre-fit"
        # path is None if figure should not be saved
        figure_path = pathlib.Path(figure_folder) / figure_name if save_figure else None

        fig = plot_model.data_mc(
            histogram_dict_list,
            total_model_unc,
            bin_edges,
            figure_path=figure_path,
            log_scale=log_scale,
            log_scale_x=log_scale_x,
            label=label,
            colors=colors,
            close_figure=close_figure,
        )
        figure_dict_list.append({"figure": fig, "region": region["Name"]})
    return figure_dict_list


def data_mc(
    model_prediction: model_utils.ModelPrediction,
    data: List[float],
    *,
    config: Optional[Dict[str, Any]] = None,
    figure_folder: Union[str, pathlib.Path] = "figures",
    log_scale: Optional[bool] = None,
    log_scale_x: bool = False,
    channels: Optional[Union[str, List[str]]] = None,
    colors: Optional[Dict[str, str]] = None,
    close_figure: bool = False,
    save_figure: bool = True,
) -> Optional[List[Dict[str, Any]]]:
    """Draws pre- and post-fit data/MC histograms for a ``pyhf`` model and data.

    The ``config`` argument is optional, but required to determine correct axis labels
    and binning. The information is not stored in the model, and default values are
    used if no ``config`` is supplied. This allows quickly plotting distributions for
    models that were not created with ``cabinetry``, and for which no config exists.

    Args:
        model_prediction (model_utils.ModelPrediction): model prediction to show
        data (List[float]): data to include in visualization, can either include auxdata
            (the auxdata is then stripped internally) or only observed yields
        config (Optional[Dict[str, Any]], optional): cabinetry configuration needed for
            binning and axis labels, defaults to None (uses a default binning and labels
            then)
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        log_scale (Optional[bool], optional): whether to use logarithmic vertical axis,
            defaults to None (automatically determine whether to use linear/log scale)
        log_scale_x (bool, optional): whether to use logarithmic horizontal axis,
            defaults to False
        channels (Optional[Union[str, List[str]]], optional): name of channel to show,
            or list of names to include, defaults to None (uses all channels)
        colors (Optional[Dict[str, str]], optional): map of sample names and colors to
            use in plot, defaults to None (uses default colors)
        close_figure (bool, optional): whether to close each figure, defaults to False
            (enable when producing many figures to avoid memory issues, prevents
            automatic rendering in notebooks)
        save_figure (bool, optional): whether to save figures, defaults to True

    Raises:
        ValueError: if color specification is incomplete

    Returns:
        Optional[List[Dict[str, Any]]]: list of dictionaries, where each dictionary
            contains a figure and the associated region name, or None if no figure was
            produced
    """
    # strip off auxdata (if needed) and obtain data indexed by channel (and bin)
    data_yields = model_utils._data_per_channel(model_prediction.model, data)

    # if custom colors are provided, ensure that they cover all samples
    if colors is not None:
        c_missing = set(model_prediction.model.config.samples).difference(colors.keys())
        if c_missing:
            raise ValueError(
                f"colors need to be provided for all samples, missing for {c_missing}"
            )

    # channels to include in plot, with optional filtering applied
    filtered_channels = model_utils._filter_channels(model_prediction.model, channels)

    if filtered_channels == []:
        # nothing to include in plots, warning already raised via _filter_channels
        return None

    # indices of included channels
    channel_indices = [
        model_prediction.model.config.channels.index(ch) for ch in filtered_channels
    ]

    # process channel by channel
    figure_dict_list = []
    for i_chan, channel_name in zip(channel_indices, filtered_channels):
        histogram_dict_list = []  # one dict per region/channel

        if config is not None:
            # get the region dictionary from the config for binning / variable name
            # for histogram inputs this information is not available, so need to fall
            # back to defaults
            region_dict = configuration.region_dict(config, channel_name)
            if region_dict.get("Binning", None) is not None:
                bin_edges = builder._binning(region_dict)
            else:
                bin_edges = np.arange(len(data_yields[i_chan]) + 1)
            variable = region_dict.get("Variable", "bin")
        else:
            # fall back to defaults if no config is specified
            bin_edges = np.arange(len(data_yields[i_chan]) + 1)
            variable = "bin"

        for i_sam, sample_name in enumerate(model_prediction.model.config.samples):
            histogram_dict_list.append(
                {
                    "label": sample_name,
                    "isData": False,
                    "yields": model_prediction.model_yields[i_chan][i_sam],
                    "variable": variable,
                }
            )

        # add data sample
        histogram_dict_list.append(
            {
                "label": "Data",
                "isData": True,
                "yields": data_yields[i_chan],
                "variable": variable,
            }
        )

        # path is None if figure should not be saved
        figure_path = (
            pathlib.Path(figure_folder)
            / _figure_name(channel_name, model_prediction.label)
            if save_figure
            else None
        )
        label = f"{channel_name}\n{model_prediction.label}"

        fig = plot_model.data_mc(
            histogram_dict_list,
            np.asarray(model_prediction.total_stdev_model_bins[i_chan][-1]),
            bin_edges,
            figure_path=figure_path,
            log_scale=log_scale,
            log_scale_x=log_scale_x,
            label=label,
            colors=colors,
            close_figure=close_figure,
        )
        figure_dict_list.append({"figure": fig, "region": channel_name})
    return figure_dict_list


def templates(
    config: Dict[str, Any],
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    close_figure: bool = False,
    save_figure: bool = True,
) -> List[Dict[str, Any]]:
    """Visualizes template histograms (after post-processing) for systematic variations.

    The original template histogram for systematic variations (before post-processing)
    is also included in the visualization.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        close_figure (bool, optional): whether to close each figure, defaults to False
            (enable when producing many figures to avoid memory issues, prevents
            automatic rendering in notebooks)
        save_figure (bool, optional): whether to save figures, defaults to True

    Returns:
        List[Dict[str, Any]]: list of dictionaries, where each dictionary contains a
            figure and the associated region / sample / systematic names
    """
    log.info("visualizing systematic templates")
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    figure_folder = pathlib.Path(figure_folder) / "templates"

    # could do this via the route module instead
    figure_dict_list = []
    for region in config["Regions"]:
        for sample in config["Samples"]:
            if sample.get("Data", False):
                # skip data
                continue

            # loop over systematics (if they exist)
            for systematic in config.get("Systematics", []):
                histo_name = (
                    region["Name"]
                    + "_"
                    + sample["Name"]
                    + "_"
                    + systematic["Name"]
                    + "*_modified*"
                ).replace(" ", "-")
                # create a list of paths to histograms matching the pattern
                variation_paths = [
                    pathlib.Path(h_name)
                    for h_name in glob.glob(str(histogram_folder / histo_name))
                ]
                # only keep up/down variations, and sort alphabetically
                # (sorting to have consistent order, and simplified debugging)
                variation_paths = sorted(
                    v for v in variation_paths if ("Up" in v.name or "Down" in v.name)
                )

                if len(variation_paths) == 0:
                    # no associated templates (normalization systematics)
                    continue

                # extract nominal histogram
                nominal_histo = histo.Histogram.from_config(
                    histogram_folder, region, sample, {}
                )
                bins = nominal_histo.bins
                # variable is a required config setting for ntuple inputs, but not for
                # histogram inputs, so default to "observable" for these cases (the
                # actual bin edges are preserved here, so not using "bin" as in data_mc)
                variable = region.get("Variable", "observable")
                nominal = {"yields": nominal_histo.yields, "stdev": nominal_histo.stdev}

                # extract original and modified (after post-processing) variation
                # histograms, if they exist
                up_orig = {}
                down_orig = {}
                up_mod = {}
                down_mod = {}
                for variation_path in variation_paths:
                    # original variation, before post-processing
                    variation_path_orig = pathlib.Path(
                        str(variation_path).replace("_modified", "")
                    )
                    var_histo_orig = histo.Histogram.from_path(variation_path_orig)
                    var_orig = {
                        "yields": var_histo_orig.yields,
                        "stdev": var_histo_orig.stdev,
                    }

                    # variation after post-processing
                    var_histo_mod = histo.Histogram.from_path(variation_path)
                    var_mod = {
                        "yields": var_histo_mod.yields,
                        "stdev": var_histo_mod.stdev,
                    }
                    if "Up" in variation_path.parts[-1]:
                        up_orig.update(var_orig)
                        up_mod.update(var_mod)
                    else:
                        down_orig.update(var_orig)
                        down_mod.update(var_mod)

                figure_label = (
                    f"region: {region['Name']}\nsample: {sample['Name']}"
                    f"\nsystematic: {systematic['Name']}"
                )
                figure_name = (
                    f"{region['Name']}_{sample['Name']}_{systematic['Name']}.pdf"
                ).replace(" ", "-")
                figure_path = figure_folder / figure_name if save_figure else None

                fig = plot_model.templates(
                    nominal,
                    up_orig,
                    down_orig,
                    up_mod,
                    down_mod,
                    bins,
                    variable,
                    figure_path=figure_path,
                    label=figure_label,
                    close_figure=close_figure,
                )
                figure_dict_list.append(
                    {
                        "figure": fig,
                        "region": region["Name"],
                        "sample": sample["Name"],
                        "systematic": systematic["Name"],
                    }
                )
    return figure_dict_list


def correlation_matrix(
    fit_results: fit.FitResults,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    pruning_threshold: float = 0.0,
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Draws a correlation matrix.

    Args:
        fit_results (fit.FitResults): fit results, including correlation matrix and
            parameter labels
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        pruning_threshold (float, optional): minimum correlation for a parameter to
            have with any other parameters to not get pruned, defaults to 0.0
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True

    Returns:
        matplotlib.figure.Figure: the correlation matrix figure
    """
    # path is None if figure should not be saved
    figure_path = (
        pathlib.Path(figure_folder) / "correlation_matrix.pdf" if save_figure else None
    )

    # create a matrix that is True if a correlation is below threshold, and True on the
    # diagonal
    below_threshold = np.where(
        np.abs(fit_results.corr_mat) < pruning_threshold, True, False
    )
    np.fill_diagonal(below_threshold, True)
    # get list of booleans specifying if everything in rows/columns is below threshold
    all_below_threshold = np.all(below_threshold, axis=0)
    # get list of booleans specifying if rows/columns correspond to fixed parameter
    # (0 correlations)
    fixed_parameter = np.all(np.equal(fit_results.corr_mat, 0.0), axis=0)
    # get indices of rows/columns where everything is below threshold, or the parameter
    # is fixed
    delete_indices = np.where(np.logical_or(all_below_threshold, fixed_parameter))
    # delete rows and columns where all correlations are below threshold / parameter is
    # fixed
    corr_mat = np.delete(
        np.delete(fit_results.corr_mat, delete_indices, axis=1), delete_indices, axis=0
    )
    labels = np.delete(fit_results.labels, delete_indices)

    fig = plot_result.correlation_matrix(
        corr_mat, labels, figure_path=figure_path, close_figure=close_figure
    )
    return fig


def pulls(
    fit_results: fit.FitResults,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    exclude: Optional[Union[str, List[str], Tuple[str, ...]]] = None,
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Draws a pull plot of parameter results and uncertainties.

    Args:
        fit_results (fit.FitResults): fit results, including correlation matrix and
            parameter labels
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        exclude (Optional[Union[str, List[str], Tuple[str, ...]]], optional): parameter
            or parameters to exclude from plot, defaults to None (nothing excluded)
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True

    Returns:
        matplotlib.figure.Figure: the pull figure
    """
    # path is None if figure should not be saved
    figure_path = pathlib.Path(figure_folder) / "pulls.pdf" if save_figure else None
    labels_np = np.asarray(fit_results.labels)

    if exclude is None:
        exclude_set = set()
    elif isinstance(exclude, str):
        exclude_set = {exclude}
    else:
        exclude_set = set(exclude)

    # exclude fixed parameters from pull plot
    exclude_set.update(
        [
            label
            for i_np, label in enumerate(labels_np)
            if fit_results.uncertainty[i_np] == 0.0
        ]
    )

    # exclude staterror parameters from pull plot (they are centered at 1)
    exclude_set.update([label for label in labels_np if label[0:10] == "staterror_"])

    # filter out user-specified parameters
    mask = [True if label not in exclude_set else False for label in labels_np]
    bestfit = fit_results.bestfit[mask]
    uncertainty = fit_results.uncertainty[mask]
    labels_np = labels_np[mask]

    fig = plot_result.pulls(
        bestfit,
        uncertainty,
        labels_np,
        figure_path=figure_path,
        close_figure=close_figure,
    )
    return fig


def ranking(
    ranking_results: fit.RankingResults,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    max_pars: Optional[int] = None,
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Produces a ranking plot showing the impact of parameters on the POI.

    The parameters are shown in decreasing order of greatest post-fit impact.

    Args:
        ranking_results (fit.RankingResults): fit results, and pre- and post-fit impacts
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        max_pars (Optional[int], optional): number of parameters to include, defaults to
            None (which means all parameters are included)
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True

    Returns:
        matplotlib.figure.Figure: the ranking figure
    """
    # path is None if figure should not be saved
    figure_path = pathlib.Path(figure_folder) / "ranking.pdf" if save_figure else None

    # sort parameters by decreasing maximum post-fit impact
    max_postfit_impact = np.maximum(
        np.abs(ranking_results.postfit_up), np.abs(ranking_results.postfit_down)
    )

    # get indices to sort by decreasing impact
    sorted_indices = np.argsort(max_postfit_impact)[::-1]
    bestfit = ranking_results.bestfit[sorted_indices]
    uncertainty = ranking_results.uncertainty[sorted_indices]
    labels = np.asarray(ranking_results.labels)[sorted_indices]  # labels are list
    prefit_up = ranking_results.prefit_up[sorted_indices]
    prefit_down = ranking_results.prefit_down[sorted_indices]
    postfit_up = ranking_results.postfit_up[sorted_indices]
    postfit_down = ranking_results.postfit_down[sorted_indices]

    if max_pars is not None:
        # only keep leading parameters in ranking
        bestfit = bestfit[:max_pars]
        uncertainty = uncertainty[:max_pars]
        labels = labels[:max_pars]
        prefit_up = prefit_up[:max_pars]
        prefit_down = prefit_down[:max_pars]
        postfit_up = postfit_up[:max_pars]
        postfit_down = postfit_down[:max_pars]

    fig = plot_result.ranking(
        bestfit,
        uncertainty,
        labels,
        prefit_up,
        prefit_down,
        postfit_up,
        postfit_down,
        figure_path=figure_path,
        close_figure=close_figure,
    )
    return fig


def scan(
    scan_results: fit.ScanResults,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Visualizes the results of a likelihood scan.

    Args:
        scan_results (fit.ScanResults): results of a likelihood scan
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True

    Returns:
        matplotlib.figure.Figure: the likelihood scan figure
    """
    # replace [], needed for staterrors
    figure_name = (
        "scan_" + scan_results.name.replace("[", "_").replace("]", "") + ".pdf"
    )
    # path is None if figure should not be saved
    figure_path = pathlib.Path(figure_folder) / figure_name if save_figure else None

    fig = plot_result.scan(
        scan_results.name,
        scan_results.bestfit,
        scan_results.uncertainty,
        scan_results.parameter_values,
        scan_results.delta_nlls,
        figure_path=figure_path,
        close_figure=close_figure,
    )
    return fig


def limit(
    limit_results: fit.LimitResults,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Visualizes observed and expected CLs values as a function of the POI.

    Args:
        limit_results (fit.LimitResults): results of upper limit determination
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True

    Returns:
        matplotlib.figure.Figure: the CLs figure
    """
    # path is None if figure should not be saved
    figure_path = pathlib.Path(figure_folder) / "limit.pdf" if save_figure else None

    fig = plot_result.limit(
        limit_results.observed_CLs,
        limit_results.expected_CLs,
        limit_results.poi_values,
        1 - limit_results.confidence_level,
        figure_path=figure_path,
        close_figure=close_figure,
    )
    return fig


def modifier_grid(
    model: pyhf.pdf.Model,
    *,
    figure_folder: Union[str, pathlib.Path] = "figures",
    split_by_sample: bool = False,
    close_figure: bool = True,
    save_figure: bool = True,
) -> mpl.figure.Figure:
    """Visualizes the modifier structure of a model in a 2d grid.

    Args:
        model (pyhf.pdf.Model): model to visualize
        figure_folder (Union[str, pathlib.Path], optional): path to the folder to save
            figures in, defaults to "figures"
        split_by_sample (bool, optional): whether to use (channel, parameter) grids
            for each sample, defaults to False (if enabled, uses (sample, parameter)
            grids for each channel)
        close_figure (bool, optional): whether to close figure, defaults to True
        save_figure (bool, optional): whether to save figure, defaults to True
    """
    # collect modifier types affecting each (channel, sample, parameter) combination
    modifier_map = model_utils._modifier_map(model)

    # build 2d grids: one grid per channel or one grid per sample
    if split_by_sample:
        # one (channel, parameter) grid per sample
        axis_labels = [
            model.config.samples,
            model.config.channels,
            model.config.par_order,
        ]
    else:
        # one (sample, parameter) grid per channel
        axis_labels = [
            model.config.channels,
            model.config.samples,
            model.config.par_order,
        ]

    grid_list = [
        np.zeros(shape=(len(axis_labels[1]), len(axis_labels[2])))
        for _ in axis_labels[0]
    ]

    category_to_int_map = {
        "normfactor": 0,
        "shapefactor": 1,
        "shapesys": 2,
        "lumi": 3,
        "staterror": 4,
        "normsys + histosys": 5,
        "histosys": 6,
        "normsys": 7,
        "none": 8,
    }

    # fill the list of grids with information about which modifiers enter where
    for i_grid, grid_label in enumerate(axis_labels[0]):
        for j_axis, axis_label in enumerate(axis_labels[1]):
            # extract channel and sample names depending on grid formatting
            if split_by_sample:
                chan, sam = axis_label, grid_label  # one grid per sample
            else:
                chan, sam = grid_label, axis_label  # one grid per channel

            for k_par, par in enumerate(axis_labels[2]):
                modifiers = modifier_map[(chan, sam, par)]
                # assign integer value to field depending on modifiers found
                if modifiers == []:
                    value = category_to_int_map["none"]
                else:
                    try:
                        # look up value from category map
                        value = category_to_int_map[" + ".join(sorted(modifiers)[::-1])]
                    except KeyError:
                        log.error(
                            f"modifiers for {chan}, {sam}, {par} not supported:"
                            f" {modifiers}"
                        )
                        raise
                grid_list[i_grid][j_axis, k_par] = value

    # translation from value to category label for plotting
    int_to_category_map = {label: val for val, label in category_to_int_map.items()}

    # path is None if figure should not be saved
    figure_path = (
        pathlib.Path(figure_folder) / "modifier_grid.pdf" if save_figure else None
    )

    fig = plot_model.modifier_grid(
        grid_list,
        axis_labels,
        int_to_category_map,
        figure_path=figure_path,
        close_figure=close_figure,
    )
    return fig



"""Provides features to apply functions to template histograms."""

import fnmatch
import logging
from typing import Any, Callable, Dict, List, Literal, Optional

import boost_histogram as bh

from cabinetry import configuration


log = logging.getLogger(__name__)


# type of a function processing templates, takes sample-region-systematic-template,
# returns None
# template can be "Up" / "Down" for variations, or None for nominal
ProcessorFunc = Callable[
    [Dict[str, Any], Dict[str, Any], Dict[str, Any], Optional[Literal["Up", "Down"]]],
    None,
]

# type of a user-defined function for template processing, takes sample-region-
# systematic-template, returns a boost_histogram.Histogram
# template can be any string (to match "Up" / "Down"), or None / "*" to match nominal
UserTemplateFunc = Callable[
    [Dict[str, Any], Dict[str, Any], Dict[str, Any], Optional[str]], bh.Histogram
]

# type of a function called with names of region-sample-systematic-template,
# which returns either a ProcessorFunc or None (in case of no match)
# the template argument is None for the nominal case, and "Up" / "Down" otherwise
MatchFunc = Callable[
    [str, str, str, Optional[Literal["Up", "Down"]]], Optional[ProcessorFunc]
]

# type of wrapper function that that turns a user-defined template processing function
# (which returns a histogram) into a function that returns None
WrapperFunc = Callable[[UserTemplateFunc], ProcessorFunc]


class Router:
    """Holds user-defined processing functions and matches functions to templates.

    Provides functions for matching a pattern to apply the right function to each
    template.

    Attributes:
        template_builders (List[Dict[str, Any]]): user-defined processors for template
            building
        template_builder_wrapper (Optional[WrapperFunc]): wrapper to apply on user-
            defined template builders
    """

    def __init__(self) -> None:
        """Initialize a Router instance, with no processors or wrappers defined."""
        # initialize all lists of processor types the user can specify
        self.template_builders: List[Dict[str, Any]] = []

        # initialize the wrapper to be used for template building to turn the user-
        # defined function (which returns a histogram) into one that saves the histogram
        # this wrapper needs to be set before using _find_template_builder_match to
        # properly handle the user-defined template builder
        self.template_builder_wrapper: Optional[WrapperFunc] = None

    @staticmethod
    def _register_processor(
        processor_list: List[Dict[str, Any]],
        region_name: str,
        sample_name: str,
        systematic_name: str,
        template: Optional[str],
    ) -> Callable[[UserTemplateFunc], UserTemplateFunc]:
        """Decorator for registering a custom processor function.

        The function is added to the list provided as function argument. Currently this
        function is used only for template builder functions, but could be used for
        additional types of functions as well. This requires extending the return type
        accordingly.

        Args:
            region_name (str): name of the region to apply the function to, or "*" to
                apply to all regions
            sample_name (str): name of the sample to apply the function to, or "*" to
                apply to all samples
            systematic_name (str): name of the systematic to apply the function to, or
               "*" to apply to all systematics
            template (Optional[str]): name of the template to apply the function to, "*"
                to apply to all templates, or None to apply to nominal only

        Returns:
            Callable[[UserTemplateFunc], UserTemplateFunc]: the function to
            register a processor
        """

        def _register(func: UserTemplateFunc) -> UserTemplateFunc:
            """Registers a processor function to be applied when matching a pattern.

            The pattern is specified by region-sample-systematic-template. To support
            functions other than template builder functions, the argument and return
            types need to be extended.

            Args:
                func (UserTemplateFunc): the function to register
            """
            processor_list.append(
                {
                    "region": region_name,
                    "sample": sample_name,
                    "systematic": systematic_name,
                    "template": template,
                    "name": func.__name__,
                    "func": func,
                }
            )
            return func

        return _register

    def register_template_builder(
        self,
        *,
        region_name: str = "*",
        sample_name: str = "*",
        systematic_name: str = "*",
        template: Optional[str] = "*",
    ) -> Callable[[UserTemplateFunc], UserTemplateFunc]:
        """Decorator for registering a template builder function.

        The function is added to the list stored in the ``template_builders`` member
        variable.

        Args:
            region_name (str, optional): name of the region to apply the function to,
                defaults to "*" (apply to all regions)
            sample_name (str, optional): name of the sample to apply the function to,
                defaults to "*" (apply to all samples)
            systematic_name (str, optional): name of the systematic to apply the
                function to, defaults to "*" (apply to all systematics)
            template (Optional[str], optional): name of the template to apply the
                function to (e.g. "Up" or "Down"), or None to apply to nominal only,
                defaults to "*" (apply to all templates, including nominal)

        Returns:
            Callable[[UserTemplateFunc], UserTemplateFunc]: the function to register a
            processor
        """
        return self._register_processor(
            self.template_builders, region_name, sample_name, systematic_name, template
        )

    @staticmethod
    def _find_match(
        processor_list: List[Dict[str, Any]],
        region_name: str,
        sample_name: str,
        systematic_name: str,
        template: Optional[Literal["Up", "Down"]],
    ) -> Optional[UserTemplateFunc]:
        """Returns a function matching the provided specification.

        This is currently only used for template builder functions, but could be used
        for other types of functions by extending the return type accordingly.

        Args:
            processor_list (List[Dict[str, Any]]): list of processors to search in
            region_name (str): region name
            sample_name (str): sample name
            systematic_name (str): systematic name (can use empty string for nominal)
            template (Optional[Literal["Up", "Down"]]): template name ("Up", "Down"), or
                None for nominal

        Returns:
            Optional[UserTemplateFunc]: processor function matching the description,
            or None if no matches are found
        """
        matches = []
        for processor in processor_list:
            region_matches = fnmatch.fnmatch(region_name, processor["region"])
            sample_matches = fnmatch.fnmatch(sample_name, processor["sample"])
            systematic_matches = fnmatch.fnmatch(
                systematic_name, processor["systematic"]
            )
            # template can only be None (nominal), "Up", "Down"
            # processor["template"] can be
            #   - None: apply to nominal only
            #   - "*": apply to all templates
            #   - another string: do normal matching, can never match nominal then
            if template is None:
                # nominal template is matched by processors applying to nominal
                # templates (None), and by processors applying to all templates ("*")
                template_matches = (processor["template"] is None) or (
                    processor["template"] == "*"
                )
            elif processor["template"] is None:
                # template is not None (nominal), but processor is None (nominal), so it
                # does not match
                template_matches = False
            else:
                # neither template nor processor are None, can do normal matching
                template_matches = fnmatch.fnmatch(template, processor["template"])
            if (
                region_matches
                and sample_matches
                and systematic_matches
                and template_matches
            ):
                matches.append(processor["func"])

        if len(matches) > 1:
            log.warning(
                f"found {len(matches)} matches, continuing with the first one "
                f"({matches[0].__name__})"
            )
        elif len(matches) == 0:
            return None
        return matches[0]

    def _find_template_builder_match(
        self,
        region_name: str,
        sample_name: str,
        systematic_name: str,
        template: Optional[Literal["Up", "Down"]],
    ) -> Optional[ProcessorFunc]:
        """Returns wrapped template builder function matching provided specification.

        If no matches are found, returns None. Wraps the user-defined function in the
        provided wrapper, if no wrapper is found raises an error.

        Args:
            region_name (str): region name
            sample_name (str): sample name
            systematic_name (str): systematic name (can use empty string for nominal)
            template (Optional[Literal["Up", "Down"]]): template name ("Up", "Down"), or
                None for nominal

        Raises:
            ValueError: when no template wrapper is set

        Returns:
            Optional[ProcessorFunc]: wrapped template builder function matching the
            description, or None if no matches are found
        """
        if self.template_builder_wrapper is None:
            # a wrapper needs to be defined already to convert the user-defined
            # UserTemplateFunc into a ProcessorFunc
            raise ValueError("no template builder wrapper defined")

        match = self._find_match(
            self.template_builders, region_name, sample_name, systematic_name, template
        )

        if match is not None:
            # if user-defined function was found, wrap and return it
            return self.template_builder_wrapper(match)
        return None


def apply_to_all_templates(
    config: Dict[str, Any],
    default_func: ProcessorFunc,
    *,
    match_func: Optional[MatchFunc] = None,
) -> None:
    """Applies the supplied function ``default_func`` to all templates.

    The templates are specified by the configuration file. The function takes four
    arguments in this order:

    - the dict specifying region information
    - the dict specifying sample information
    - the dict specifying systematic information
    - the template being considered: "Up", "Down", or None for the nominal template

    In addition it is possible to specify a function that returns custom overrides. If
    one is found for a given template, it is used instead of the default.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        default_func (ProcessorFunc): function to be called for every template by
            default
        match_func: (Optional[MatchFunc], optional): function that returns user-defined
            functions to override the call to ``default_func``, defaults to None (then
            it is not used)
    """
    for region in config["Regions"]:
        log.debug(f"  in region {region['Name']}")

        for sample in config["Samples"]:
            log.debug(f"    reading sample {sample['Name']}")
            # region dependence of sample is checked below via histogram_is_needed

            # loop over nominal templates and all existing systematics
            for systematic in [{}] + config.get("Systematics", []):
                # determine how many templates need to be considered
                templates: List[Optional[Literal["Up", "Down"]]]
                if systematic == {}:
                    # only nominal template is needed
                    templates = [None]
                else:
                    # systematics can have up and down template
                    templates = ["Up", "Down"]

                for template in templates:
                    # determine whether a histogram is needed for this
                    # specific combination of sample-region-systematic-template
                    # could consider caching the results of histogram_is_needed
                    histo_needed = configuration.histogram_is_needed(
                        region, sample, systematic, template
                    )

                    if not histo_needed:
                        # no further action is needed, continue with the next
                        # region-sample-systematic combination
                        continue

                    log.debug(
                        "      variation "
                        f"{systematic['Name'] if template is not None else 'Nominal'}"
                        f"{' ' + template if template is not None else ''}"
                    )

                    func_override = None
                    if match_func is not None:
                        # check whether a user-defined function was registered that
                        # matches this region-sample-systematic-template
                        systematic_name = (
                            systematic["Name"] if template is not None else ""
                        )
                        func_override = match_func(
                            region["Name"], sample["Name"], systematic_name, template
                        )
                    if func_override is not None:
                        # call the user-defined function
                        log.debug(
                            f"executing user-defined override {func_override.__name__}"
                        )
                        func_override(region, sample, systematic, template)
                    else:
                        # call the provided default function
                        default_func(region, sample, systematic, template)



"""Creates ntuples used as input for a minimal example of cabinetry."""
import os

import matplotlib.pyplot as plt
import numpy as np
import uproot


def toy_distribution(noncentral, multiplier, offset, num_events):
    return (
        np.random.noncentral_chisquare(5, noncentral, num_events) * multiplier + offset
    )


def toy_weights(total_yield, num_events):
    avg = total_yield / float(num_events)
    weights = np.random.normal(avg, avg * 0.1, num_events)
    # re-normalize to make sure sum of weights exactly matches avg
    weights *= total_yield / np.sum(weights)
    return weights


def get_samples(num_events):
    dist_s = toy_distribution(10, 12, 350, num_events)
    dist_b = toy_distribution(10, 25, 0, num_events)
    dist_b_var = toy_distribution(12, 30, 0, num_events)
    return [dist_s, dist_b, dist_b_var]


def get_weights(yield_s, yield_b, yield_b_var, num_events):
    w_s = toy_weights(yield_s, num_events)
    w_b = toy_weights(yield_b, num_events)
    w_b_var = toy_weights(yield_b_var, num_events)
    return [w_s, w_b, w_b_var]


def create_pseudodata(yield_s, yield_b):
    # create a dataset with some slightly different composition
    scale_s = 1.2
    scale_b = 1.0
    dist_s = toy_distribution(10, 12, 350, int(yield_s * scale_s))
    dist_b = toy_distribution(10, 25, 0, int(yield_b * scale_b))
    pseudodata = np.hstack((dist_s, dist_b))
    np.random.shuffle(pseudodata)
    return pseudodata


def create_lepton_charge(n_events):
    # lepton charge is +1 or -1 for all events, just to have an extra variable
    charge = (np.random.randint(0, 2, size=n_events) * 2) - 1
    return charge


def create_file(file_name, distributions, weights, labels, *, extra_weights=None):
    if extra_weights is None:
        extra_weights = []
    n_events = len(weights[0])
    with uproot.recreate(file_name) as f:
        # write the predicted processes
        for i, label in enumerate(labels):
            lep_charge = create_lepton_charge(n_events)
            if label == "background":
                f[label] = {
                    "jet_pt": distributions[i],
                    "weight": weights[i],
                    "lep_charge": lep_charge,
                    "weight_up": extra_weights[0],
                    "weight_down": extra_weights[1],
                }
            else:
                f[label] = {
                    "jet_pt": distributions[i],
                    "weight": weights[i],
                    "lep_charge": lep_charge,
                }


def create_file_pseudodata(file_name, pseudodata):
    n_events = len(pseudodata)
    with uproot.recreate(file_name) as f:
        # write pseudodata
        lep_charge = create_lepton_charge(n_events)
        f["pseudodata"] = {"jet_pt": pseudodata, "lep_charge": lep_charge}


def read_file(file_name):
    distributions = []
    weights = []
    labels = []
    with uproot.open(file_name) as f:
        all_trees = f.keys(filter_classname="TTree", recursive=True)
        for tree in all_trees:
            distributions.append(f[tree]["jet_pt"].array(library="np"))
            weights.append(f[tree]["weight"].array(library="np"))
            labels.append(f[tree].name)
    return distributions, weights, labels


def read_file_pseudodata(file_name):
    with uproot.open(file_name) as f:
        distribution = f["pseudodata"]["jet_pt"].array(library="np")
    return distribution


def plot_distributions(data, weights, labels, pseudodata, bins):
    bin_width_str = str(int(bins[1] - bins[0]))
    yield_each = [str(round(np.sum(w), 1)) for w in weights]

    # plot normalized distributions
    for i in reversed(range(len(data))):
        plt.hist(
            data[i],
            weights=weights[i],
            bins=bins,
            label=labels[i],
            histtype="step",
            density=True,
        )
    plt.legend(frameon=False)
    plt.xlabel(r"jet $p_T$ [GeV]")
    plt.ylabel("normalized")
    plt.savefig("normalized.png", dpi=200)
    plt.close()

    # plot stack
    plt.clf()
    # filter out samples with "varied" in the label,
    # which correspond to background systematic variations
    labels_with_yield = [
        labels[i] + " " + yield_each[i]
        for i in range(len(labels))
        if "varied" not in labels[i]
    ]
    data = [data[i] for i in range(len(labels)) if "varied" not in labels[i]]
    weights = [weights[i] for i in range(len(labels)) if "varied" not in labels[i]]
    pseudodata_label = "pseudodata " + str(len(pseudodata))
    plt.hist(
        data[::-1],
        weights=weights[::-1],
        bins=bins,
        label=labels_with_yield[::-1],
        histtype="stepfilled",
        stacked=True,
    )
    plt.hist(pseudodata, bins=bins, label=pseudodata_label, histtype="step", color="k")
    plt.legend(frameon=False)
    plt.xlabel(r"jet $p_T$ [GeV]")
    plt.ylabel("events / " + bin_width_str + " GeV")
    plt.savefig("stacked.png", dpi=200)
    plt.close()


def run(output_directory, *, visualize=False):
    # configuration
    num_events = 5000
    yield_s = 125
    yield_b = 1000
    yield_b_var = 1000
    labels = ["signal", "background", "background_varied"]  # names of processes
    file_name = output_directory + "/prediction.root"
    file_name_pseudodata = output_directory + "/data.root"

    np.random.seed(0)

    # distributions for two processes, plus a background uncertainty
    distributions = get_samples(num_events)

    # corresponding weights
    weights = get_weights(yield_s, yield_b, yield_b_var, num_events)

    # weights for an extra background uncertainty
    weight_var_up = (1.2 * (distributions[1] / 350)) * weights[1]
    weight_var_down = (np.ones_like(weight_var_up) * 0.7) * weights[1]

    # create a pseudodataset
    pseudodata = create_pseudodata(yield_s, yield_b)

    # write it all to a file
    create_file(
        file_name,
        distributions,
        weights,
        labels,
        extra_weights=[weight_var_up, weight_var_down],
    )
    create_file_pseudodata(file_name_pseudodata, pseudodata)

    # read the files again
    d_read, w_read, l_read = read_file(file_name)
    pd_read = read_file_pseudodata(file_name_pseudodata)

    # compare predictions from before/after reading
    np.testing.assert_allclose(d_read, distributions)
    np.testing.assert_allclose(w_read, weights)
    np.testing.assert_allclose(pd_read, pseudodata)

    if visualize:
        # visualize results
        bins = np.linspace(0, 1200, 24 + 1)
        plot_distributions(d_read, w_read, l_read, pseudodata, bins)


if __name__ == "__main__":
    output_directory = "inputs/"
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    run(output_directory)



"""Visualizes inference results with matplotlib."""

import logging
import math
import pathlib
from typing import List, Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import packaging.version

from cabinetry.visualize import utils


log = logging.getLogger(__name__)

# handling of matplotlib<3.6
if packaging.version.parse(mpl.__version__) < packaging.version.parse("3.6"):
    MPL_GEQ_36 = False  # pragma: no cover
    MPL_STYLE = "seaborn-colorblind"  # pragma: no cover
else:
    MPL_GEQ_36 = True
    MPL_STYLE = "seaborn-v0_8-colorblind"


def correlation_matrix(
    corr_mat: np.ndarray,
    labels: Union[List[str], np.ndarray],
    *,
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a correlation matrix.

    Args:
        corr_mat (np.ndarray): the correlation matrix to plot
        labels (Union[List[str], np.ndarray]): names of parameters in the correlation
            matrix
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the correlation matrix figure
    """
    # rounding for test in CI to match reference
    fig, ax = plt.subplots(
        figsize=(round(5 + len(labels) / 1.6, 1), round(3 + len(labels) / 1.6, 1)),
        dpi=100,
        layout="constrained",
    )
    im = ax.imshow(corr_mat, vmin=-1, vmax=1, cmap="RdBu")

    ax.set_xticks(np.arange(len(labels)))
    ax.set_yticks(np.arange(len(labels)))
    ax.set_xticklabels(labels, rotation=45, ha="right")
    ax.set_yticklabels(labels)

    fig.colorbar(im, ax=ax)
    ax.set_aspect("auto")  # to get colorbar aligned with matrix

    # add correlation as text
    for (j, i), corr in np.ndenumerate(corr_mat):
        text_color = "white" if abs(corr_mat[j, i]) > 0.75 else "black"
        if abs(corr) > 0.005:
            ax.text(i, j, f"{corr:.2f}", ha="center", va="center", color=text_color)

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def pulls(
    bestfit: np.ndarray,
    uncertainty: np.ndarray,
    labels: Union[List[str], np.ndarray],
    *,
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a pull plot.

    Args:
        bestfit (np.ndarray): best-fit parameter results
        uncertainty (np.ndarray): parameter uncertainties
        labels (Union[List[str], np.ndarray]): parameter names
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the pull figure
    """
    num_pars = len(bestfit)
    y_positions = np.arange(num_pars)[::-1]
    fig, ax = plt.subplots(figsize=(6, 1 + num_pars / 4), dpi=100, layout="constrained")
    ax.errorbar(bestfit, y_positions, xerr=uncertainty, fmt="o", color="black")

    ax.fill_between([-2, 2], -0.5, len(bestfit) - 0.5, color="yellow")
    ax.fill_between([-1, 1], -0.5, len(bestfit) - 0.5, color="limegreen")
    ax.vlines(0, -0.5, len(bestfit) - 0.5, linestyles="dotted", color="black")

    ax.set_xlim([-3, 3])
    ax.set_xlabel(r"$\left(\hat{\theta} - \theta_0\right) / \Delta \theta$")
    ax.set_ylim([-0.5, num_pars - 0.5])
    ax.set_yticks(y_positions)
    ax.set_yticklabels(labels)
    ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())  # minor ticks
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def ranking(
    bestfit: np.ndarray,
    uncertainty: np.ndarray,
    labels: Union[List[str], np.ndarray],
    impact_prefit_up: np.ndarray,
    impact_prefit_down: np.ndarray,
    impact_postfit_up: np.ndarray,
    impact_postfit_down: np.ndarray,
    *,
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a ranking plot.

    Args:
        bestfit (np.ndarray): best-fit parameter results
        uncertainty (np.ndarray): parameter uncertainties
        labels (Union[List[str], np.ndarray]): parameter labels
        impact_prefit_up (np.ndarray): pre-fit impact in "up" direction per parameter
        impact_prefit_down (np.ndarray): pre-fit impact in "down" direction per
            parameter
        impact_postfit_up (np.ndarray): post-fit impact in "up" direction per parameter
        impact_postfit_down (np.ndarray): post-fit impact in "down" direction per
            parameter
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the ranking figure
    """
    num_pars = len(bestfit)

    # layout to make space for legend on top
    leg_space = 1.0 / (num_pars + 3) + 0.03
    if MPL_GEQ_36:
        import matplotlib.layout_engine

        layout = matplotlib.layout_engine.ConstrainedLayoutEngine(
            rect=[0, 0, 1.0, 1 - leg_space]
        )
    else:
        layout = None  # pragma: no cover  # layout set after figure creation instead

    mpl.style.use(MPL_STYLE)
    fig, ax_pulls = plt.subplots(
        figsize=(8, 2.5 + num_pars * 0.45), dpi=100, layout=layout
    )
    ax_impact = ax_pulls.twiny()  # second x-axis with shared y-axis, used for pulls

    # since pull axis is below impact axis, flip them so pulls show up on top
    ax_pulls.set_zorder(1)  # pulls axis on top
    ax_pulls.patch.set_visible(False)  # pulls axis does not hide impact axis

    # increase font sizes
    for item in (
        [ax_pulls.yaxis.label, ax_pulls.xaxis.label, ax_impact.xaxis.label]
        + ax_pulls.get_yticklabels()
        + ax_pulls.get_xticklabels()
        + ax_impact.get_xticklabels()
    ):
        item.set_fontsize("large")

    # lines through pulls of -1, 0, 1 for orientation
    # line does not go all the way up to the top x-axis, since it
    # belongs to the bottom x-axis
    ax_pulls.vlines(
        -1, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )
    ax_pulls.vlines(
        0, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )
    ax_pulls.vlines(
        1, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )

    y_pos = np.arange(num_pars)[::-1]

    # pre-fit up
    pre_up = ax_impact.barh(
        y_pos, impact_prefit_up, fill=False, linewidth=1, edgecolor="C0"
    )
    # pre-fit down
    pre_down = ax_impact.barh(
        y_pos, impact_prefit_down, fill=False, linewidth=1, edgecolor="C5"
    )
    # post-fit up
    post_up = ax_impact.barh(y_pos, impact_postfit_up, color="C0")
    # post-fit down
    post_down = ax_impact.barh(y_pos, impact_postfit_down, color="C5")
    # nuisance parameter pulls
    pulls = ax_pulls.errorbar(bestfit, y_pos, xerr=uncertainty, fmt="o", color="k")

    ax_pulls.set_xlabel(r"$\left(\hat{\theta} - \theta_0\right) / \Delta \theta$")
    ax_impact.set_xlabel(r"$\Delta \mu$")
    ax_pulls.set_xlim([-2, 2])
    ax_impact.set_xlim([-5, 5])
    ax_pulls.set_ylim([-1, num_pars])

    # impact axis limits: need largest pre-fit impact
    impact_max = np.amax(np.fabs(np.hstack((impact_prefit_up, impact_prefit_down))))
    ax_impact.set_xlim([-impact_max * 1.1, impact_max * 1.1])

    # minor ticks
    for axis in [ax_pulls.xaxis, ax_impact.xaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax_pulls.set_yticks(y_pos)
    ax_pulls.set_yticklabels(labels)

    ax_pulls.tick_params(direction="in", which="both")
    ax_impact.tick_params(direction="in", which="both")

    fig.legend(
        (pre_up, pre_down, post_up, post_down, pulls),
        (
            r"pre-fit impact: $\theta = \hat{\theta} + \Delta \theta$",
            r"pre-fit impact: $\theta = \hat{\theta} - \Delta \theta$",
            r"post-fit impact: $\theta = \hat{\theta} + \Delta \hat{\theta}$",
            r"post-fit impact: $\theta = \hat{\theta} - \Delta \hat{\theta}$",
            "pulls",
        ),
        frameon=False,
        loc="upper left",
        ncol=3,
        fontsize="large",
    )

    if not MPL_GEQ_36:
        fig.tight_layout(rect=[0, 0, 1.0, 1 - leg_space])  # pragma: no cover

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def scan(
    par_name: str,
    par_mle: float,
    par_unc: float,
    par_vals: np.ndarray,
    par_nlls: np.ndarray,
    *,
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a figure showing the results of a likelihood scan.

    Args:
        par_name (str): name of parameter used in scan
        par_mle (float): best-fit result for parameter
        par_unc (float): best-fit parameter uncertainty
        par_vals (np.ndarray): values used in scan over parameter
        par_nlls (np.ndarray): -2 log(L) offset at each scan point
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the likelihood scan figure
    """
    mpl.style.use(MPL_STYLE)
    fig, ax = plt.subplots(layout="constrained")

    y_lim = max(par_nlls) * 1.2  # upper y-axis limit, 20% headroom

    # lines through y=1 and y=4 to show confidence intervals
    ax.plot([par_vals[0], par_vals[-1]], [1, 1], ":", color="gray")
    ax.plot([par_vals[0], par_vals[-1]], [4, 4], ":", color="gray")

    # position for text - right edge of the figure, with slight padding
    text_x_pos = par_vals[-1] - 0.01 * (par_vals[-1] - par_vals[0])
    # only draw text if it fits in the figure
    if y_lim >= 1:
        ax.text(text_x_pos, 1.0, "68% CL", ha="right", va="bottom", color="gray")
    if y_lim >= 4:
        ax.text(text_x_pos, 4.0, "95% CL", ha="right", va="bottom", color="gray")

    # Gaussian at best-fit parameter value for reference
    val_grid = np.linspace(par_vals[0], par_vals[-1], 100)
    gaussian_approx = [((par_val - par_mle) / par_unc) ** 2 for par_val in val_grid]
    ax.plot(val_grid, gaussian_approx, "--", color="C5", label="Gaussian approximation")

    # scan results
    ax.plot(par_vals, par_nlls, "-", color="C0")
    ax.plot(par_vals, par_nlls, "X", color="C0", label="parameter scan")

    # increase font sizes
    for item in (
        [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax.set_xlabel(par_name)
    ax.set_xlim(par_vals[0], par_vals[-1])
    ax.set_ylabel(r"$-2 \Delta \log(L)$")
    ax.set_ylim(0, y_lim)
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")

    ax.legend(frameon=False, fontsize="large")

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def limit(
    observed_CLs: np.ndarray,
    expected_CLs: np.ndarray,
    poi_values: np.ndarray,
    cls_target: float,
    *,
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws observed and expected CLs values as function of the parameter of interest.

    Args:
        observed_CLs (np.ndarray): observed CLs values
        expected_CLs (np.ndarray): expected CLs values, including 1 and 2 sigma bands
        poi_values (np.ndarray): parameter of interest values used in scan
        cls_target (float): target CLs value to visualize as horizontal line
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the CLs figure
    """
    fig, ax = plt.subplots(layout="constrained")

    xmin = min(poi_values)
    xmax = max(poi_values)

    # observed CLs values
    ax.plot(poi_values, observed_CLs, "o-", color="black", label=r"observed CL$_S$")

    # expected CLs
    ax.plot(
        poi_values, expected_CLs[:, 2], "--", color="black", label=r"expected CL$_S$"
    )

    # 1 and 2 sigma bands
    ax.fill_between(
        poi_values,
        expected_CLs[:, 1],
        expected_CLs[:, 3],
        color="limegreen",
        label=r"expected CL$_S$ $\pm 1\sigma$",
    )
    ax.fill_between(
        poi_values,
        expected_CLs[:, 0],
        expected_CLs[:, 4],
        color="yellow",
        label=r"expected CL$_S$ $\pm 2\sigma$",
        zorder=0,  # draw beneath 1 sigma band
    )

    # determine whether CLs value shown in percent is integer (float.is_integer() is not
    # sufficient after calculation of 1-confidence_level in fit.limit)
    cls_pct_is_integer = math.isclose(cls_target * 100, round(cls_target * 100))
    cls_label = f"{cls_target:.{0 if cls_pct_is_integer else 2}%}"
    # line through CLs = cls_target
    ax.hlines(
        cls_target,
        xmin=xmin,
        xmax=xmax,
        linestyle="dashdot",
        color="red",
        label="CL$_S$ = " + cls_label,  # 2 decimals unless they are both 0, then 0
        zorder=1,  # draw beneath observed / expected
    )

    # increase font sizes
    for item in (
        [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax.legend(frameon=False, fontsize="large")

    ax.set_xlabel(r"$\mu$")
    ax.set_ylabel(r"$\mathrm{CL}_{s}$")
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([0, 1])
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")

    utils._save_and_close(fig, figure_path, close_figure)
    return fig



import logging

import numpy as np
import pytest

from cabinetry import smooth


@pytest.mark.parametrize(
    "orig, mod",
    [
        (([1, 3, 2, 5], 4), [1, 2, 3, 5]),
        (([1, 3, 2, 5, 11, 3, 8], 7), [1, 2, 3, 5, 5, 8, 8]),
    ],
)
def test__medians_353(orig, mod):
    smooth._medians_353(orig[0], orig[1])
    assert np.allclose(orig[0], mod)


def test_smooth_353qh_twice(caplog):
    caplog.set_level(logging.DEBUG)
    hist_smooth = smooth.smooth_353qh_twice([1, 3, 2, 5])
    assert isinstance(hist_smooth, list)
    assert np.allclose(hist_smooth, [1, 2, 3.25, 5])
    assert np.allclose(
        smooth.smooth_353qh_twice([1, 3, 2, 5, 11, 3, 8]),
        [1.0, 2.125, 3.66666675, 5.08333349, 6.16666651, 7.375, 8.0],
    )

    # with np.ndarray
    hist_smooth = smooth.smooth_353qh_twice(np.asarray([1, 3, 2, 5]))
    assert isinstance(hist_smooth, np.ndarray)
    assert hist_smooth.dtype == float
    assert np.allclose(hist_smooth, [1, 2, 3.25, 5])
    caplog.clear()

    assert np.allclose(smooth.smooth_353qh_twice([1, 3]), [1, 3])
    assert "at least three points needed for smoothing, no smoothing applied" in [
        rec.message for rec in caplog.records
    ]

    assert np.allclose(
        smooth.smooth_353qh_twice(
            [0.16505027, 0.92460960, 0.04765633, 0.36997846, 0.37872216]
        ),
        [0.16505027, 0.21628231, 0.31874642, 0.36997846, 0.36997846],
    )

    assert np.allclose(
        smooth.smooth_353qh_twice(
            [0.29093575, 0.93474650, 0.29723760, 0.61081952, -1.00000000]
        ),
        [0.29093575, 0.52901053, 0.47264278, -0.06587567, -1.00000000],
    )
    caplog.clear()



"""Provides containers for inference results."""

from typing import Dict, List, NamedTuple, Tuple

import numpy as np


class FitResults(NamedTuple):
    """Collects fit results in one object.

    Args:
        bestfit (np.ndarray): best-fit results of parameters
        uncertainty (np.ndarray): uncertainties of best-fit parameter results, evaluated
            with Hessian
        labels (List[str]): parameter labels
        corr_mat (np.ndarray): parameter correlation matrix
        best_twice_nll (float): -2 log(likelihood) at best-fit point
        goodess_of_fit (float, optional): goodness-of-fit p-value, defaults to -1
        minos_uncertainty (Dict[str, Tuple[float, float]]): uncertainties of best-fit
            parameter results indexed by parameter name, calculated with MINOS
    """

    bestfit: np.ndarray
    uncertainty: np.ndarray
    labels: List[str]
    corr_mat: np.ndarray
    best_twice_nll: float
    goodness_of_fit: float = -1
    minos_uncertainty: Dict[str, Tuple[float, float]] = {}


class RankingResults(NamedTuple):
    """Collects nuisance parameter ranking results in one object.

    The best-fit results per parameter, the uncertainties, and the labels should not
    include the parameter of interest, since no impact for it is calculated.

    Args:
        bestfit (np.ndarray): best-fit results of parameters
        uncertainty (np.ndarray): uncertainties of best-fit parameter results
        labels (List[str]): parameter labels
        prefit_up (np.ndarray): pre-fit impact in "up" direction
        prefit_down (np.ndarray): pre-fit impact in "down" direction
        postfit_up (np.ndarray): post-fit impact in "up" direction
        postfit_down (np.ndarray): post-fit impact in "down" direction
    """

    bestfit: np.ndarray
    uncertainty: np.ndarray
    labels: List[str]
    prefit_up: np.ndarray
    prefit_down: np.ndarray
    postfit_up: np.ndarray
    postfit_down: np.ndarray


class ScanResults(NamedTuple):
    """Collects likelihood scan results in one object.

    Args:
        name (str): name of parameter in scan
        bestfit (float): best-fit parameter value from unconstrained fit
        uncertainty (float): uncertainty of parameter in unconstrained fit
        parameter_values (np.ndarray): parameter values used in scan
        delta_nlls (np.ndarray): -2 log(L) difference at each scanned point
    """

    name: str
    bestfit: float
    uncertainty: float
    parameter_values: np.ndarray
    delta_nlls: np.ndarray


class LimitResults(NamedTuple):
    """Collects parameter upper limit results in one object.

    Args:
        observed_limit (float): observed limit
        expected_limit (np.ndarray): expected limit, including 1 and 2 sigma bands
        observed_CLs (np.ndarray): observed CLs values
        expected_CLs (np.ndarray): expected CLs values, including 1 and 2 sigma bands
        poi_values (np.ndarray): POI values used in scan
        confidence_level (float): confidence level used for parameter limits
    """

    observed_limit: float
    expected_limit: np.ndarray
    observed_CLs: np.ndarray
    expected_CLs: np.ndarray
    poi_values: np.ndarray
    confidence_level: float


class SignificanceResults(NamedTuple):
    """Collects results from a discovery significance calculation in one object.

    Args:
        observed_p_value (float): observed p-value
        observed_significance (float): observed significance
        expected_p_value (float): expected/median p-value
        expected_significance (float): expected/median significance
    """

    observed_p_value: float
    observed_significance: float
    expected_p_value: float
    expected_significance: float
